# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=

import json
import os
import tempfile

import vmdb


class AnsiblePlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(AnsibleStepRunner())


class AnsibleStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {
            "ansible": str,
            "playbook": str,
            "group": "image",
            "tags": "all",
            "config_file": "",
            "extra_vars": {},
        }

    def run(self, values, settings, state):
        tag = values["ansible"]
        playbook = values["playbook"]
        ansible_tags = values["tags"]
        group_name = values["group"]
        config_file = values["config_file"]
        extra_vars = values["extra_vars"]
        mount_point = state.tags.get_builder_mount_point(tag)
        rootfs_tarball = settings["rootfs-tarball"]

        inventory = self.create_inventory(mount_point, group_name)
        if not hasattr(state, "ansible_inventory"):
            state.ansible_inventory = []
        state.ansible_inventory.append(inventory)
        vmdb.progress(f"Created {inventory} for Ansible inventory")

        extra_vars["rootfs_tarball"] = rootfs_tarball
        if not hasattr(state, "ansible_vars_file"):
            state.ansible_vars_file = []

        env = dict(os.environ)
        env["ANSIBLE_NOCOWS"] = "1"
        if config_file:
            if os.path.exists(config_file):
                env["ANSIBLE_CONFIG"] = config_file
                vmdb.progress(f"Using Ansible config file {config_file}")
            else:
                raise RuntimeError(f"Ansible config file {config_file} does not exist")
        extra_vars_json = json.dumps(extra_vars)
        vmdb.runcmd(
            [
                "ansible-playbook",
                "-c",
                "chroot",
                "-i",
                inventory,
                "--tags",
                ansible_tags,
                "-e",
                extra_vars_json,
                playbook,
            ],
            env=env,
        )

    def teardown(self, values, settings, state):
        if hasattr(state, "ansible_inventory"):
            self.remove(state.ansible_inventory)

    def remove(self, filenames):
        for filename in filenames:
            if os.path.exists(filename):
                vmdb.progress("Removing {}".format(filename))
                os.remove(filename)

    def create_inventory(self, chroot, group_name):
        fd, filename = tempfile.mkstemp()
        os.write(fd, f"[{group_name}]\n{chroot}\n".encode())
        os.close(fd)
        return filename



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=

import logging
import os
import tempfile

import vmdb


class MountPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(MountStepRunner())


class MountStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {"mount": str, "dirname": "", "mount-on": "", "zerofree": True, "keep_resolvconf": False}

    def run(self, values, settings, state):
        self.mount_rootfs(values, settings, state)

    def teardown(self, values, settings, state):
        self.unmount_rootfs(values, settings, state)
        tag = values["mount"]
        run_zerofree = values.get("zerofree", True)
        if run_zerofree:
            fstype = state.tags.get_fstype(tag)
            if fstype.startswith("ext"):
                dev = state.tags.get_dev(tag)
                vmdb.runcmd(["zerofree", dev])
            else:
                logging.info(f"not running zerofree on {tag} with {fstype} filesystem")

    def mount_rootfs(self, values, settings, state):
        tag = values["mount"]
        dirname = values["dirname"] or None
        mount_on = values["mount-on"] or None

        device = state.tags.get_dev(tag)

        if dirname:
            if not mount_on:
                raise Exception("no mount-on tag given")

            if not state.tags.has_tag(mount_on):
                raise Exception("cannot find tag {}".format(mount_on))

            mount_point = os.path.normpath(
                os.path.join(
                    state.tags.get_builder_mount_point(mount_on), "./" + dirname
                )
            )

            if not os.path.exists(mount_point):
                os.makedirs(mount_point)
        else:
            dirname = "/"
            mount_point = tempfile.mkdtemp()

        vmdb.runcmd(["mount", device, mount_point])
        state.tags.set_builder_mount_point(tag, mount_point, cached=True)
        state.tags.set_target_mount_point(tag, dirname)

        # detect the filesystem and uuid if they are not already known. This
        # will be the case if the tags were defined via the kpartx plugin,
        # rather than by the mkfs plugin
        if not state.tags.get_fstype(tag):
            fstype = (
                vmdb.runcmd(["blkid", "-c/dev/null", "-ovalue", "-sTYPE", device])
                .decode()
                .strip()
            )
            state.tags.set_fstype(tag, fstype)
            uuid = state.tags.get_fsuuid(tag)
            if not uuid:
                uuid = (
                    vmdb.runcmd(["blkid", "-c/dev/null", "-ovalue", "-sUUID", device])
                    .decode()
                    .strip()
                )
                state.tags.set_fsuuid(tag, uuid)

        return mount_point

    def unmount_rootfs(self, values, settings, state):
        tag = values["mount"]
        mount_point = state.tags.get_builder_mount_point(tag)
        if mount_point is None:
            return

        keep_resolvconf = values.get("keep_resolvconf", False)

        resolv_conf = os.path.join(mount_point, "etc/resolv.conf")
        if os.path.exists(resolv_conf) and not keep_resolvconf:
            os.remove(resolv_conf)

        try:
            vmdb.unmount(mount_point)
        except vmdb.NotMounted as e:
            logging.warning(str(e))

        if not values["mount-on"]:
            os.rmdir(mount_point)



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=
import json
import os
import re

import vmdb


class KpartxPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(KpartxStepRunner())


class KpartxStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {"kpartx": str, "tags": []}

    def run(self, values, settings, state):
        device = values["kpartx"]
        tags = values["tags"]
        if not tags:
            tags = state.tags.get_tags()
        else:
            # add the tags defined in the plugin spec
            for tag in tags:
                state.tags.append(tag)

        devs = self.kpartx(device)
        for tag, dev in zip(tags, devs):
            vmdb.progress(f"remembering {dev} as {tag}")
            state.tags.set_dev(tag, dev)

    def kpartx(self, device):
        # Allocate the loop device first -- creating one if required -- before
        # mapping the partitions usign kpartx. This handles scenarios where no
        # free loop devices are available.
        loop_device = (
            vmdb.runcmd(["losetup", "-f", "--show", device]).decode("UTF-8").strip()
        )
        output = vmdb.runcmd(["kpartx", "-usv", loop_device]).decode("UTF-8")
        for line in output.splitlines():
            words = line.split()
            if words[0] == "add":
                name = words[2]
                yield "/dev/mapper/{}".format(name)

    def teardown(self, values, settings, state):
        device = values["kpartx"]
        vmdb.runcmd(["kpartx", "-dsv", device])
        # docker containers on macOS don't honor the kpartx cleanup command, so
        # we have to do the clean up ourselves... :-/
        loop_devs = set()
        for tag in state.tags.get_tags():
            dev = state.tags.get_dev(tag)
            m = re.match(r"^/dev/mapper/(?P<loop>.*)p\d+$", dev)
            if m is not None:
                if os.path.exists(dev):
                    vmdb.runcmd(["dmsetup", "-v", "remove", dev])
                loop = m.group("loop")
                loop_devs.add("/dev/{}".format(loop))

        for loop_dev in loop_devs:
            # Check if loop device is actually active:
            loopdev_info = json.loads(
                vmdb.runcmd(["losetup", "--json", "-l", loop_dev])
            )["loopdevices"]
            if len(loopdev_info) != 1:
                # Something is wrong, this must not happen
                raise RuntimeError("losetup returned more than one device")
            loopdev_info = loopdev_info.pop()
            if loopdev_info["back-file"] is None:
                continue
            vmdb.progress(
                "Loop device {} is still bound to back-file {}, "
                "removing loop device".format(loop_dev, loopdev_info["back-file"])
            )
            vmdb.runcmd(["losetup", "-d", loop_dev])



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=

import os
import shutil
import subprocess

import vmdb


class DebootstrapPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(DebootstrapStepRunner())


class DebootstrapStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {
            "debootstrap": str,
            "target": str,
            "mirror": str,
            "arch": "",
            "keyring": "",
            "install_keyring": False,
            "variant": "-",
            "components": ["main"],
            "include": [],
            "exclude": [],
            "require_empty_target": True,
            "tls_ca_certs": [],
        }

    def run(self, values, settings, state):
        suite = values["debootstrap"]
        tag = values["target"]
        target = state.tags.get_builder_mount_point(tag)
        mirror = values["mirror"]
        keyring = values["keyring"] or None
        install_keyring = values["install_keyring"]
        tls_ca_certs = values["tls_ca_certs"]
        include = values["include"]
        exclude = values["exclude"]
        require_empty = values["require_empty_target"]
        arch = values["arch"] or state.arch
        state.arch = arch
        vmdb.progress(f"debootstrap plugin set architecture to {state.arch}")
        variant = values["variant"]
        components = values["components"]

        if not (suite and tag and target and mirror):
            raise Exception("missing arg for debootstrap step")

        if os.path.exists(target) and require_empty:
            allowed_names = ["lost+found"]
            names = [n for n in os.listdir(target) if n not in allowed_names]
            if len(names) > 0:
                raise Exception(
                    f"debootstrap target {target} is a not an empty directory: {names}"
                )

        bad_certs = [c for c in tls_ca_certs if not c.endswith(".crt")]
        if bad_certs:
            raise RuntimeError(f'TLS cert(s) do not have a ".crt" suffix: {bad_certs}')

        cmd = [
            "debootstrap",
            "--arch",
            arch,
            "--variant",
            variant,
            "--components",
            ",".join(components),
        ]

        remove_pkgs = []
        if keyring:
            cmd.extend(["--keyring", keyring])
            if install_keyring and "gnupg" not in include:
                include.append("gnupg")
                # If gnupg needed to be installed it should be removed again to
                # minimize the installation footprint
                remove_pkgs.append("gnupg")

        if include:
            cmd.extend(["--include", ",".join(include)])

        if exclude:
            cmd.extend(["--exclude", ",".join(exclude)])

        cmd.extend([suite, target, mirror])

        vmdb.runcmd(cmd)

        if keyring and install_keyring:
            keyring_basename = os.path.basename(keyring)
            chroot_keyring = os.path.join(target, keyring_basename)
            shutil.copyfile(keyring, os.path.join(target, keyring_basename))
            vmdb.runcmd_chroot(target, ["apt-key", "add", f"/{keyring_basename}"])
            os.remove(chroot_keyring)

        if tls_ca_certs:
            for ca_cert in tls_ca_certs:
                target_cert_path = os.path.join(
                    target,
                    "usr/local/share/ca-certificates",
                    os.path.basename(ca_cert),
                )

                shutil.copyfile(ca_cert, target_cert_path)
                vmdb.progress(f"Copied {ca_cert} -> {target_cert_path}")
            vmdb.runcmd_chroot(target, ["update-ca-certificates"])

        if remove_pkgs:
            vmdb.runcmd_chroot(
                target,
                [
                    "apt-get",
                    "remove",
                    "--purge",
                    "-y",
                ]
                + remove_pkgs,
            )

    def run_even_if_skipped(self, values, settings, state):
        tag = values["target"]
        target = state.tags.get_builder_mount_point(tag)
        state.arch = values["arch"] or state.arch
        vmdb.progress(f"debootstrap plugin set architecture to {state.arch}")
        vmdb.runcmd_chroot(target, ["apt-get", "update"])



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import vmdb


class MkimgPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(MkimgStepRunner())


class MkimgStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {"mkimg": str, "size": str}

    def run(self, values, settings, state):
        filename = values["mkimg"]
        size = values["size"]

        if not isinstance(filename, str):
            raise vmdb.NotString("mkimg", filename)
        if not filename:
            raise vmdb.IsEmptyString("mkimg", filename)

        if not isinstance(size, str):
            raise vmdb.NotString("mkimg: size", size)

        vmdb.runcmd(["qemu-img", "create", "-f", "raw", filename, size])



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=

import io
import logging
import os
import selectors
import subprocess
import sys

_verbose = False


def set_verbose_progress(verbose):
    global _verbose
    _verbose = verbose


def error(msg):
    logging.error(msg, exc_info=True)
    sys.stderr.write("ERROR: {}\n".format(msg))
    sys.stderr.flush()


def progress(msg):
    logging.info(msg)
    if _verbose:
        sys.stdout.write("{}\n".format(msg))
        sys.stdout.flush()


def _log_line(line, stream_out, stream_label):
    if line:
        line = line.decode("UTF8")
        if _verbose:
            stream_out.write(line)
            stream_out.flush()
        logging.debug("%s: %s", stream_label, line)
    return bool(line)


def runcmd(argv, **kwargs):
    progress("Exec: %r" % (argv, ))
    env = kwargs.get("env", os.environ.copy())
    env["LC_ALL"] = "C.UTF8"
    kwargs["env"] = env
    kwargs["stdout"] = kwargs.get("stdout", subprocess.PIPE)
    kwargs["stderr"] = kwargs.get("stderr", subprocess.PIPE)
    for name in env:
        logging.debug(f"ENV: {name}={env[name]}")

    stdout_buffer = io.BytesIO()

    def _log_stdout(stream_in):
        line = stream_in.readline()
        stdout_buffer.write(line)
        return _log_line(line, sys.stdout, "STDOUT")

    def _log_stderr(stream_in):
        line = stream_in.readline()
        return _log_line(line, sys.stderr, "STDERR")

    selector = selectors.DefaultSelector()
    p = subprocess.Popen(argv, **kwargs)
    selector.register(p.stdout, selectors.EVENT_READ, _log_stdout)
    selector.register(p.stderr, selectors.EVENT_READ, _log_stderr)

    while p.poll() is None:
        events = selector.select()
        for key, _ in events:
            callback = key.data
            callback(key.fileobj)
    p.wait()
    selector.close()

    # drain stdout and stderr
    while _log_stdout(p.stdout) or _log_stderr(p.stderr):
        pass

    if p.returncode != 0:
        raise RuncmdError("Program failed: {}".format(p.returncode))
    return stdout_buffer.getvalue()


def runcmd_chroot(chroot, argv, *argvs, **kwargs):
    full_argv = ["chroot", chroot] + argv
    return runcmd(full_argv, *argvs, **kwargs)


def _procdir(chroot):
    proc = os.path.join(chroot, "proc")
    if not os.path.exists(proc):
        os.mkdir(proc, mode=0o755)
    return proc


def _log_stdout(data):
    logging.debug("STDOUT: %r", data)
    return data


def _log_stderr(data):
    logging.debug("STDERR: %r", data)
    return data


class RuncmdError(Exception):
    pass



# Copyright 2018  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


# Unmount a directory, including any mount points under that
# directory. If /mnt/foo is given, and /mnt/foo/bar is also mounted,
# unmount /mnt/foo/bar first, and /mnt/foo then. Look for sub-mounts
# in /proc/mounts.


import logging
import subprocess


def unmount(what, mounts=None, real_unmount=None):
    logging.debug("Unmounting {} and everything on top of it".format(what))
    if mounts is None:  # pragma: no cover
        mounts = _read_proc_mounts()
    if real_unmount is None:  # pragma: no cover
        real_unmount = _real_unmount

    mounts = _parse_proc_mounts(mounts)
    dirnames = _find_what_to_unmount(mounts, what)
    for dirname in dirnames:
        real_unmount(dirname)
    logging.debug("Finished unmounting {}".format(what))


def _read_proc_mounts():  # pragma: no cover
    with open("/proc/mounts") as f:
        return f.read()


def _real_unmount(what):  # pragma: no cover
    try:
        subprocess.check_call(["umount", what])
    except subprocess.CalledProcessError as e:
        logging.info("unmount failed, but ignoring that: %s", e)


def _parse_proc_mounts(text):
    return [line.split()[:2] for line in text.splitlines()]


def _find_what_to_unmount(mounts, what):
    dirname = _find_mount_point(mounts, what)
    dirnameslash = dirname + "/"
    to_unmount = [
        point
        for dev, point in mounts
        if point == dirname or point.startswith(dirnameslash)
    ]
    return list(reversed(sorted(to_unmount)))


def _find_mount_point(mounts, what):
    for dev, point in mounts:
        if what in (dev, point):
            return point
    raise NotMounted(what)


class NotMounted(Exception):
    def __init__(self, what):
        super().__init__("Not mounted: {}".format(what))



# Copyright 2023  Andy Piper
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=

import json

import vmdb


class LvscanPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(LvscanStepRunner())


class LvscanStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {"lvscan": str, "tags": []}

    def run(self, values, settings, state):
        for v in get_logical_volumes(values):
            dev = v["lv_path"]
            tag = v["lv_name"]
            vmdb.progress(f"remembering {dev} as {tag}")
            state.tags.append(tag)
            state.tags.set_dev(tag, dev)
            vmdb.runcmd(["lvchange", "-a", "y", dev])

    def teardown(self, values, settings, state):
        vgname = values["lvscan"]
        vmdb.runcmd(["vgchange", "-an", vgname])


def get_logical_volumes(values):
    vgname = values["lvscan"]
    tags = values["tags"]
    vmdb.runcmd(["lvscan"])
    volumes = json.loads(
        vmdb.runcmd(
            [
                "lvdisplay",
                "-C",
                "-o",
                "lv_name,vg_name,lv_path",
                "--report-format",
                "json",
            ]
        )
    )["report"][0]["lv"]

    # ensure all tags have a corresponding volume, but not all volumes need
    # a matching tag
    matched_vols = [
        v for v in volumes if v["lv_name"] in tags and v["vg_name"] == vgname
    ]
    if len(matched_vols) < len(tags):
        matched_tags = [v["lv_name"] for v in matched_vols]
        unmatched_tags = sorted(set(tags).difference(matched_tags))
        raise RuntimeError(
            f'unmatched tags in volume group "{vgname}": {unmatched_tags}'
        )
    return matched_vols



# Copyright 2022  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import logging
import os
import shutil
import tempfile

import vmdb


class CryptsetupPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(CryptsetupStepRunner())


class CryptsetupStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {
            "cryptsetup": str,
            "name": str,
            "password": "",
            "key-file": "",
            "key-cmd": "",
        }

    def run(self, step, settings, state):
        underlying = step["cryptsetup"]
        crypt_name = step["name"]
        password = step["password"] or None
        key_file = step["key-file"] or None
        key_cmd = step["key-cmd"] or None

        if not isinstance(underlying, str):
            raise vmdb.NotString("cryptsetup", underlying)

        if not isinstance(crypt_name, str):
            raise vmdb.NotString("cryptsetup: tag", crypt_name)

        if password is None and key_file is None and key_cmd is None:
            raise Exception(
                "cryptsetup step MUST define one of password, key-file, or key-cmd"
            )

        if password is not None and key_file is not None:
            raise Exception(
                "cryptsetup step MUST define only one of password or key-file"
            )

        if password is not None and key_cmd is not None:
            raise Exception(
                "cryptsetup step MUST define only one of password or key-cmd"
            )

        if key_file is not None and key_cmd is not None:
            raise Exception(
                "cryptsetup step MUST define only one of key_file or key-cmd"
            )

        state.tmp_key_file = None
        rmtmp = False

        if password is not None:
            key_file = self._write_temp(password)
            rmtmp = True

        if key_cmd is not None:
            output = vmdb.runcmd(["sh", "-ec", key_cmd])
            output = output.decode("UTF-8")
            key = output.splitlines()[0]
            key_file = self._write_temp(key)
            rmtmp = True

        assert key_file is not None

        dev = state.tags.get_dev(underlying)
        if dev is None:
            for t in state.tags.get_tags():
                logging.debug(
                    "tag %r dev %r mp %r",
                    t,
                    state.tags.get_dev(t),
                    state.tags.get_builder_mount_point(t),
                )
            assert 0

        vmdb.runcmd(
            [
                "cryptsetup",
                "luksFormat",
                "--batch-mode",
                "--type=luks2",
                "--pbkdf=argon2id",
                dev,
                key_file,
            ]
        )
        vmdb.runcmd(
            [
                "cryptsetup",
                "open",
                "--key-file",
                key_file,
                "--allow-discards",
                dev,
                crypt_name,
            ]
        )

        crypt_dev = "/dev/mapper/{}".format(crypt_name)
        assert os.path.exists(crypt_dev)

        uuid = vmdb.runcmd(["cryptsetup", "luksUUID", dev]).decode("UTF8").strip()

        state.tags.append(crypt_name)
        state.tags.set_dev(crypt_name, crypt_dev)
        state.tags.set_luksuuid(crypt_name, uuid)
        state.tags.set_dm(crypt_name, crypt_name)

        vmdb.progress(
            f"LUKS: name={crypt_name} dev={crypt_dev} luksuuid={uuid} dm={crypt_name}"
        )
        vmdb.progress(f"LUKS: {state.tags._tags}")
        vmdb.progress("remembering LUKS device {} as {}".format(crypt_dev, crypt_name))

        if rmtmp:
            os.remove(key_file)

    def _write_temp(self, password):
        fd, filename = tempfile.mkstemp()
        os.close(fd)
        open(filename, "w").write(password)
        return filename

    def teardown(self, step, settings, state):
        x = state.tmp_key_file
        if x is not None and os.path.exists(x):
            os.remove(x)

        crypt_name = step["name"]

        crypt_dev = "/dev/mapper/{}".format(crypt_name)
        vmdb.runcmd(["cryptsetup", "close", crypt_dev])



# Copyright 2020  Lars Wirzenius, Peter Lawler
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import os

import vmdb


SCRIPT = """\
#!/bin/sh

rootpart="$(findmnt -n -o SOURCE /)"
rootdev="/dev/$(lsblk -no pkname "$rootpart")"
partno="$(lsblk -n --list "$rootpart" | wc -l)"

flock $rootdev sfdisk -f $rootdev -N "$partno" <<EOF
,+
EOF

sleep 5

udevadm settle

sleep 5

flock $rootdev partprobe $rootdev

mount -o remount,rw $rootpart

resize2fs -f $rootpart

exit 0
"""

SERVICE = """\
[Unit]
Description=resize root file system
Before=local-fs-pre.target
DefaultDependencies=no

[Service]
Type=oneshot
TimeoutSec=infinity
ExecStart=/usr/sbin/resize-rootfs
ExecStart=/bin/systemctl --no-reload disable %n

[Install]
RequiredBy=local-fs-pre.target
"""


class ResizeRootfsPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(ResizeRootfsStepRunner())


class ResizeRootfsStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {"resize-rootfs": str}

    def run(self, values, settings, state):
        tag = values["resize-rootfs"]
        root = state.tags.get_builder_mount_point(tag)

        partprobe = self.relative(root, "/sbin/parted")
        if not os.path.exists(partprobe):
            raise Exception("partprobe is not installed on image, can't resize")

        script_path = "/usr/sbin/resize-rootfs"
        service_path = "/etc/systemd/system/resize-rootfs.service"
        requires_path = "/etc/systemd/system/systemd-remount-fs.service.requires/"

        # Install the script that actually does the resizing.
        self.install(SCRIPT, root, script_path, 0o755)

        # Install the systemd service file to invoke the script at boot.
        self.install(SERVICE, root, service_path, 0o644)

        # Tell systemd to run the script at the right time. This corresponds to
        # "systemctl enable".
        self.mkdir(root, "etc/systemd/system/systemd-remount-fs.service.requires")
        self.symlink(root, service_path, requires_path)

    def relative(self, root, pathname):
        return os.path.join(root, "./" + pathname)

    def install(self, content, root, filename, mode):
        filename = self.relative(root, filename)
        with open(filename, "w") as f:
            f.write(content)
        os.chmod(filename, mode)

    def mkdir(self, root, pathname):
        pathname = self.relative(root, pathname)
        os.makedirs(pathname, exist_ok=True)

    def symlink(self, root, linkto, linkname):
        linkname = self.relative(root, linkname)
        basename = os.path.basename(linkto)
        linkname = os.path.join(linkname, basename)
        os.symlink(linkto, linkname)



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import logging
import os

import vmdb


class UnpackRootFSPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(UnpackCacheStepRunner())


class UnpackCacheStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {"unpack-rootfs": str}

    def run(self, values, settings, state):
        fs_tag = values["unpack-rootfs"]
        rootdir = state.tags.get_builder_mount_point(fs_tag)
        logging.debug(f"settings: {settings}")
        tar_path = settings["rootfs-tarball"]
        logging.debug(f"tar_path: {tar_path!r}")
        if not tar_path:
            raise Exception(
                "--rootfs-tarball MUST be set if using the 'unpack-rootfs' step"
            )
        if os.path.exists(tar_path):
            vmdb.runcmd(["tar", "-C", rootdir, "-xf", tar_path, "--numeric-owner"])
            state.rootfs_unpacked = True



# Copyright 2020  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import importlib.util
import os


class Plugin:
    def __init__(self, app):
        self.app = app


def find_plugins(dirname, namesuffix):
    suffix = "_plugin.py"
    filenames = _find_suffixed_files(dirname, suffix)
    plugins = []
    for filename in filenames:
        for c in _classes_in_file(filename, namesuffix):
            plugins.append(c)

    return plugins


def _find_suffixed_files(dirname, suffix):
    return [
        os.path.join(dirname, basename)
        for basename in os.listdir(dirname)
        if basename.endswith(suffix)
    ]


def _classes_in_file(filename, namesuffix):
    module_name = os.path.basename(filename)[: -len(".py")]
    spec = importlib.util.spec_from_file_location(module_name, filename)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    classes = []
    for name in dir(module):
        if name.endswith(namesuffix):
            classes.append(module.__getattribute__(name))
    return classes



# Copyright 2020  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import os
import shutil
import tempfile
import unittest


import vmdb


class PluginTests(unittest.TestCase):
    def test_sets_app(self):
        plugin = vmdb.Plugin("foo")
        self.assertEqual(plugin.app, "foo")


class FindPluginsTests(unittest.TestCase):
    def setUp(self):
        self.dirname = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.dirname)

    def tests_finds_no_plugins_in_empty_directory(self):
        self.assertEqual(vmdb.find_plugins(self.dirname, ""), [])

    def tests_finds_no_plugins_when_there_are_other_files(self):
        open(os.path.join(self.dirname, "file.py"), "w").close()
        self.assertEqual(vmdb.find_plugins(self.dirname, ""), [])

    def tests_finds_no_plugin_when_file_has_none(self):
        open(os.path.join(self.dirname, "foo_plugin.py"), "w").close()
        self.assertEqual(vmdb.find_plugins(self.dirname, "Plugin"), [])

    def tests_finds_plugin_when_there_is_one(self):
        with open(os.path.join(self.dirname, "foo_plugin.py"), "w") as f:
            f.write(
                """
class FooPlugin:
   pass
"""
            )

        plugins = vmdb.find_plugins(self.dirname, "Plugin")
        self.assertEqual(len(plugins), 1)



# Copyright 2019 Gunnar Wolf
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=

import vmdb
import os
import logging
import shutil


class CopyDirPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(CopyDirStepRunner())


class CopyDirStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {
            "copy-dir": str,
            "src": str,
            "perm": 0o755,
            "umask": 0o022,
            "uid": 0,
            "gid": 0,
            "user": "",
            "group": ""
        }

    def run(self, values, settings, state):
        root = state.tags.get_builder_from_target_mount_point("/")
        newdir = values["copy-dir"]
        src = values["src"]
        perm = values["perm"]
        umask = values["umask"]

        if values["user"]:
            uid = self._get_id_from_file("/".join([root, "etc/passwd"]),
                                         values["user"])
        else:
            uid = values["uid"]

        if values["group"]:
            gid = self._get_id_from_file("/".join([root, "etc/group"]),
                                         values["group"])
        else:
            gid = values["gid"]

        tgt_dir = "/".join([root, newdir])

        # ensure src and target paths don't end with a '/'
        if src.endswith('/'):
            src = src[:-1]
        if tgt_dir.endswith('/'):
            tgt_dir = tgt_dir[:-1]

        vmdb.progress(
            "Copying dir %s to %s, uid %d, gid %d, umask %o"
            % (src, tgt_dir, uid, gid, umask)
        )

        # create missing parent dirs as required
        tgt_parent_dir = tgt_dir
        missing_parent_dirs = []
        while not os.path.exists(tgt_parent_dir):
            missing_parent_dirs.append(tgt_parent_dir)
            tgt_parent_dir = os.path.dirname(tgt_parent_dir)

        for d in reversed(missing_parent_dirs):
            self._create_dir(d, perm & ~umask, uid, gid)

        for base_dir, dirs, files in os.walk(src):
            base_tgt = base_dir.replace(src, tgt_dir)
            for dir in dirs:
                src_mode = os.stat(os.path.join(base_dir, dir)).st_mode
                umasked_mode = src_mode & ~umask
                dir_path = os.path.join(base_tgt, dir)
                if not os.path.exists(dir_path):
                    self._create_dir(dir_path, umasked_mode, uid, gid)

            for f in files:
                src_file = os.path.join(base_dir, f)
                tgt_file = os.path.join(base_tgt, f)
                src_mode = os.stat(src_file).st_mode
                umasked_mode = src_mode & ~umask
                with open(src_file, "rb") as inp:
                    with open(tgt_file, "wb") as output:
                        os.chown(tgt_file, uid, gid)
                        os.chmod(tgt_file, umasked_mode)
                        shutil.copyfileobj(inp, output)

    def _create_dir(self, dir_path, mode, uid, gid):
        os.makedirs(dir_path, mode=mode, exist_ok=True)
        os.chown(dir_path, uid, gid)

    def _get_id_from_file(self, file_path, name):
        with open(file_path) as fh:
            for line in fh:
                if line.startswith(f"{name}:"):
                    id = line.split(':', 4)[2]
                    return int(id)
        raise RuntimeError(f'Failed to find entry for "{name}" in {file_path}')


# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import argparse
import logging
import os
import sys

import vmdb


class Vmdb2:
    def __init__(self, version):
        self._version = version

    def run(self):
        args = self.parse_command_line()
        vmdb.set_verbose_progress(args.verbose)
        tvars = self.template_vars_from_args(args)

        cmd = None
        if args.image or args.output:
            if args.image:
                cmd = ReuseImageCommand(args.image, args.specfile, args.rootfs_tarball)
            else:
                cmd = NewImageCommand(args.output, args.specfile, args.rootfs_tarball)
            builder = cmd.builder()
            builder.add_template_vars(tvars)

        if cmd is None:
            sys.exit("I don't know what to do, re-run with --help")

        if args.log:
            self.setup_logging(args.log)
            self.log_startup()
        try:
            cmd.run()
        except Exception as e:
            logging.error(f"ERROR: {e}", exc_info=True)
            sys.exit(f"ERROR: {e}\n")
        logging.debug("Ending, all OK")

    def parse_command_line(self):
        p = argparse.ArgumentParser(
            prog="vmdb2", description="build disk images with Debian installed"
        )

        p.add_argument("--image", metavar="FILE")
        p.add_argument("--output", metavar="FILE")
        p.add_argument("--rootfs-tarball", metavar="FILE")
        p.add_argument("-v", "--verbose", action="store_true")
        p.add_argument("--log")
        p.add_argument("--version", action="version", version=vmdb.__version__)
        p.add_argument(
            "--variable",
            action="append",
            default=[],
            help="Specify the value for a Jinja2 template variable in the "
            'build spec file in the format "name=value". This parameter can '
            "be specified multiple times to define multiple variables.",
        )
        p.add_argument("specfile")

        return p.parse_args()

    def setup_logging(self, filename):
        fmt = "%(asctime)s %(levelname)s %(message)s"
        datefmt = "%Y-%m-%d %H:%M:%S"
        formatter = logging.Formatter(fmt, datefmt)

        handler = logging.FileHandler(filename)
        handler.setFormatter(formatter)

        logger = logging.getLogger()
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)

    def log_startup(self):
        logging.info(f"Starting vmdb2 version {self._version}")

    def template_vars_from_args(self, args):
        vars = {
            "image": args.image,
            "output": args.output,
            "rootfs_tarball": args.rootfs_tarball,
        }

        for var_spec in args.variable:
            name, value = var_spec.split("=")
            if name in vars:
                raise RuntimeError(f'Duplicate definition of variable "{name}"')

            # Remove quoes from the value if required
            if value.startswith("'"):
                value = value.strip("'")
            elif value.startswith('"'):
                value = value.strip('"')

            vars[name] = value

        return vars


class Command:
    def run(self):
        raise NotImplementedError()


class VersionCommand(Command):
    def __init__(self, version):
        self.version = version

    def run(self):
        sys.stdout.write("{}\n".format(self.version))


class NewImageCommand(Command):
    def __init__(self, filename, specfile, tarball):
        self._builder = ImageBuilder(filename, specfile, tarball)

    def builder(self):
        return self._builder

    def run(self):
        # FIXME: create file
        self._builder.build()


class ReuseImageCommand(Command):
    def __init__(self, filename, specfile, tarball):
        self._builder = ImageBuilder(filename, specfile, tarball)

    def builder(self):
        return self._builder

    def run(self):
        self._builder.build()


class ImageBuilder:
    def __init__(self, filename, specfile, tarball):
        self._filename = filename
        self._specfile = specfile
        self._tarball = tarball
        self._tvars = {}

    def add_template_vars(self, tvars):
        self._tvars.update(tvars)

    def build(self):
        spec = self.load_spec_file(self._specfile)
        state = vmdb.State()
        state.tags = vmdb.Tags()
        state.arch = (
            vmdb.runcmd(["dpkg", "--print-architecture"]).decode("UTF-8").strip()
        )
        logging.debug(f"Set state.arch={state.arch!r}")
        self.add_template_vars(state.as_dict())
        steps = spec.get_steps(self._tvars)

        # Check that we have step runners for each step
        self.load_step_runners()
        for step in steps:
            self.step_runners.find(step)

        steps_taken, core_meltdown = self.run_steps(steps, state)
        if core_meltdown:
            vmdb.progress("Something went wrong, cleaning up!")
            self.run_teardowns(steps_taken, state)
        else:
            self.run_teardowns(steps_taken, state)
            vmdb.progress("All went fine.")

        if core_meltdown:
            logging.error("An error occurred, exiting")
            sys.exit(1)

    def load_spec_file(self, filename):
        spec = vmdb.Spec()
        if filename == "-":
            vmdb.progress("Load spec from stdin")
            spec.load_file(sys.stdin)
        else:
            vmdb.progress("Load spec file {}".format(filename))
            with open(filename) as f:
                spec.load_file(f)
        vmdb.progress(f"Spec:\n{spec.as_dict()}")
        return spec

    def load_step_runners(self):
        self.step_runners = vmdb.StepRunnerList()
        plugindir = os.path.join(os.path.dirname(vmdb.__file__), "plugins")
        for klass in vmdb.find_plugins(plugindir, "Plugin"):
            klass(self).enable()

    def run_steps(self, steps, state):
        return self.run_steps_helper(steps, state, "Running step: %r", "run", False)

    def run_teardowns(self, steps, state):
        return self.run_steps_helper(
            list(reversed(steps)), state, "Running teardown: %r", "teardown", True
        )

    def run_steps_helper(self, steps, state, msg, method_name, keep_going):
        core_meltdown = False
        steps_taken = []
        settings = {"rootfs-tarball": self._tarball}

        even_if_skipped = method_name + "_even_if_skipped"
        for step in steps:
            try:
                logging.info(msg, step)
                steps_taken.append(step)
                runner = self.step_runners.find(step)
                if runner.skip(step, settings, state):
                    logging.info("Skipping as requested by unless")
                    method_names = [even_if_skipped]
                else:
                    method_names = [method_name, even_if_skipped]

                methods = [
                    getattr(runner, name)
                    for name in method_names
                    if hasattr(runner, name)
                ]

                values = runner.get_values(step)
                for method in methods:
                    logging.info("Calling %s", method)
                    method(values, settings, state)
            except KeyError as e:
                vmdb.error("Key error: %s" % str(e))
                vmdb.error(repr(e))
                core_meltdown = True
                if not keep_going:
                    break
            except BaseException as e:
                vmdb.error(str(e))
                vmdb.error(repr(e))
                core_meltdown = True
                if not keep_going:
                    break

        return steps_taken, core_meltdown



# Copyright 2019 Gunnar Wolf
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=

import vmdb
import os
import logging


class CreateDirPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(CreateDirStepRunner())


class CreateDirStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {"create-dir": str, "perm": 0o755, "uid": 0, "gid": 0}

    def run(self, values, settings, state):
        root = state.tags.get_builder_from_target_mount_point("/")
        newdir = values["create-dir"]
        path = "/".join([root, newdir])
        perm = values["perm"]
        uid = values["uid"]
        gid = values["gid"]

        logging.info(
            "Creating directory %s, uid %d, gid %d, perms %o" % (path, uid, gid, perm)
        )

        os.makedirs(path, perm)
        os.chown(path, uid, gid)



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import os

import vmdb


class CacheRootFSPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(MakeCacheStepRunner())


class MakeCacheStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {"cache-rootfs": str, "options": "--one-file-system", "force": False}

    def run(self, values, settings, state):
        fs_tag = values["cache-rootfs"]
        rootdir = state.tags.get_builder_mount_point(fs_tag)
        tar_path = settings["rootfs-tarball"]
        opts = values["options"].split()
        if not tar_path:
            raise Exception("--rootfs-tarball MUST be set")
        dirs = self._find_cacheable_mount_points(state.tags, rootdir)

        tags = state.tags
        for tag in tags.get_tags():
            vmdb.progress(
                "tag {} mounted {} cached {}".format(
                    tag, tags.get_builder_mount_point(tag), tags.is_cached(tag)
                )
            )

        vmdb.progress("caching rootdir {}".format(rootdir))
        vmdb.progress("caching relative {}".format(dirs))
        if values["force"] or not os.path.exists(tar_path):
            vmdb.runcmd(["tar"] + opts + ["-C", rootdir, "-caf", tar_path] + dirs)

    def _find_cacheable_mount_points(self, tags, rootdir):
        return list(
            sorted(
                self._make_relative(rootdir, tags.get_builder_mount_point(tag))
                for tag in tags.get_tags()
                if tags.is_cached(tag)
            )
        )

    def _make_relative(self, rootdir, dirname):
        assert dirname == rootdir or dirname.startswith(rootdir + "/")
        if dirname == rootdir:
            return "."
        return dirname[len(rootdir) + 1 :]



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


# Installing GRUB onto a disk image is a bit of a black art. I haven't
# found any good documentation for it. This plugin is written based on
# de-ciphering the build_openstack_image script. Here is an explanation
# of what I _THINK_ is happening.
#
# The crucial command is grub-install. It needs a ton of options to
# work correctly: see below in the code for the list, and the manpage
# for an explanation of what each of them means. We will be running
# grub-install in a chroot so that we use the version in the Debian
# version we're installing, rather than the host system, which might
# be any Debian version.
#
# To run grub-install in a chroot, we need to set up the chroot in
# various ways. Firstly, we need to tell grub-install which device
# file the image has. We can't just give it the image file itself,
# since it isn't inside the chroot, so instead we arrange to have a
# loop block device that covers the whole image file, and we bind
# mount /dev into the chroot so the device is available.
#
# grub-install seems to also require /proc and /sys so we bind mount
# /sys into the chroot as well. /proc is already mounted otherwise.
#
# We install the UEFI version of GRUB, and for that we additionally
# bind mount the EFI partition in the image. Oh yeah, you MUST have
# one.
#
# We also make sure the right GRUB package is installed in the chroot,
# before we run grub-install.
#
# Further, there's some configuration tweaking we need to do. See the
# code. Don't ask me why they're necessary.
#
# For cleanliness, we also undo any bind mounts into the chroot. Don't
# want to leave them in case they cause trouble.
#
# Note that this is currently assuming that UEFI and either the amd64
# (a.k.a. x86_64) or arm64 (a.k.a. aarch64) architectures are being
# used. These should probably not be hardcoded. Patch welcome.

# To use this plugin: write steps to create a root filesystem, and an
# VFAT filesystem to be mounted as /boot/efi. Install Debian onto the
# root filesystem. Then install grub with a step like this:
#
#         - grub: uefi
#           tag: root-part
#           efi: efi-part
#
# Here: "tag" is the tag for the root filesystem (and corresponding
# partition), and efi is tag for the EFI partition.
#
# The grub step will take of the rest.

import json
import logging
import os
import re

import vmdb


class GrubPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(GrubStepRunner())


class GrubStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {
            "grub": str,
            "root-fs": "",
            "efi": "",
            "efi-part": "",
            "prep": "",
            "console": "",
            "tag": "",
            "image-dev": "",
            "quiet": False,
            "kernel-params": ["biosdevname=0", "net.ifnames=0", "consoleblank=0", "rw"],
            "timeout": 0,
        }

    def run(self, values, settings, state):
        state.grub_mounts = []
        flavor = values["grub"]
        if flavor == "uefi":
            self.install_uefi(values, settings, state)
        elif flavor == "bios":
            self.install_bios(values, settings, state)
        elif flavor == "ieee1275":
            self.install_ieee1275(values, settings, state)
        else:
            raise Exception("Unknown GRUB flavor {}".format(flavor))

    def grub_uefi_variant(self, state):
        variants = {
            "amd64": ("grub-efi-amd64", "x86_64-efi"),
            "i386": ("grub-efi-ia32", "i386-efi"),
            "arm64": ("grub-efi-arm64", "arm64-efi"),
            "armhf": ("grub-efi-arm", "arm-efi"),
        }
        logging.debug(f"grub plugin: state.arch={state.arch!r}")
        try:
            return variants[state.arch]
        except KeyError:
            raise Exception(
                'GRUB UEFI package and target for "{}" unknown'.format(state.arch)
            )

    def install_uefi(self, values, settings, state):
        efi = values["efi"] or None
        efi_part = values["efi-part"] or None
        if efi is None and efi_part is None:
            raise Exception('"efi" or "efi-part" required in UEFI GRUB installation')

        vmdb.progress("Installing GRUB for UEFI")
        (grub_package, grub_target) = self.grub_uefi_variant(state)
        self.install_grub(values, settings, state, grub_package, grub_target)

    def install_bios(self, values, settings, state):
        vmdb.progress("Installing GRUB for BIOS")
        grub_package = "grub-pc"
        grub_target = "i386-pc"
        self.install_grub(values, settings, state, grub_package, grub_target)

    def grub_ieee1275_variant(self, state):
        variants = {
            "amd64": "i386",
            "ppc64": "powerpc",
            "ppc64el": "powerpc",
            "sparc": "sparc64",
        }
        logging.debug(f"grub plugin: state.arch={state.arch!r}")
        return variants.get(state.arch, state.arch)

    def install_ieee1275(self, values, settings, state):
        vmdb.progress("Installing GRUB for IEEE1275")
        grub_package = "grub-ieee1275"
        grub_target = f"{self.grub_ieee1275_variant(state)}-ieee1275"
        self.install_grub(values, settings, state, grub_package, grub_target)

    def install_grub(self, values, settings, state, grub_package, grub_target):
        console = values["console"] or None

        tag = values["tag"] or values["root-fs"] or None
        root_dev = state.tags.get_dev(tag)
        chroot = state.tags.get_builder_mount_point(tag)
        self.bind_mount_many(chroot, ["/dev", "/sys", "/proc"], state)

        image_dev = values["image-dev"] or None
        if image_dev is None:
            image_dev = self.get_image_loop_device(root_dev, chroot)

        efi = values["efi"] or None
        efi_part = values["efi-part"] or None
        if efi is not None:
            efi_dev = state.tags.get_dev(efi)
        elif efi_part is not None:
            efi_dev = state.tags.get_dev(efi_part)
        else:
            efi_dev = None

        prep = values["prep"] or None
        if prep:
            prep_dev = state.tags.get_dev(prep)
        else:
            prep_dev = None

        quiet = values["quiet"]

        if efi_dev:
            pn = efi_dev[-1]
            vmdb.runcmd(["parted", "-s", image_dev, "set", pn, "esp", "on"])
            self.mount(chroot, efi_dev, "/boot/efi", state)
        elif prep_dev:
            pn = prep_dev[-1]
            vmdb.runcmd(["parted", "-s", image_dev, "set", pn, "prep", "on"])
            image_dev = prep_dev
        self.install_package(chroot, grub_package)

        kernel_params = values["kernel-params"]

        if console == "serial":
            if "ppc64" in state.arch:
                kernel_params.extend(
                    ["loglevel=3", "console=tty0", "console=hvc0,115200n8"]
                )
            elif "arm" in state.arch:
                kernel_params.extend(
                    ["loglevel=3", "console=tty0", "console=ttyAMA0,115200n8"]
                )
            else:
                kernel_params.extend(
                    ["loglevel=3", "console=tty0", "console=ttyS0,115200n8"]
                )

        if quiet:
            kernel_params.extend(
                [
                    "quiet",
                    "systemd.show_status=false",
                    "rd.systemd.show_status=false",
                ]
            )
        else:
            kernel_params.extend(
                [
                    "systemd.show_status=true",
                ]
            )

        self.set_grub_cmdline_config(chroot, kernel_params)
        self.add_grub_crypto_disk(chroot)
        self.set_grub_timeout(chroot, values["timeout"])
        if console == "serial":
            self.add_grub_serial_console(chroot)

        vmdb.runcmd_chroot(chroot, ["grub-mkconfig", "-o", "/boot/grub/grub.cfg"])
        grub_cmd = [
            "grub-install",
            "--target=" + grub_target,
            "--no-nvram",
            "--no-floppy",
            "--modules=part_msdos part_gpt",
            "--grub-mkdevicemap=/boot/grub/device.map",
        ]
        help_out = vmdb.runcmd_chroot(chroot, ["grub-install", "--help"])
        if b"--force-extra-removable" in help_out:
            grub_cmd.append("--force-extra-removable")
        grub_cmd.append(image_dev)
        vmdb.runcmd_chroot(chroot, grub_cmd)

    #        self.unmount(state)

    def teardown(self, values, settings, state):
        self.unmount(state)

    def unmount(self, state):
        mounts = getattr(state, "grub_mounts", [])
        mounts.reverse()
        while mounts:
            mount_point = mounts.pop()
            try:
                vmdb.unmount(mount_point)
            except vmdb.NotMounted as e:
                logging.warning(str(e))

    def get_image_loop_device(self, partition_device, chroot):
        # We get /dev/mappers/loopXpY and return /dev/loopX
        m = re.match(r"^/dev/mapper/(?P<loop>.*)p\d+$", partition_device)
        if m is not None:
            loop = m.group("loop")
            return "/dev/{}".format(loop)

        # Check if the rootfs is a LVM volume
        m = re.match(r"^/dev/(?P<vgname>[^/]+)/.*$", partition_device)
        if m is not None:
            vgname = m.group("vgname")
            logging.debug(f"extracted vgname={vgname} from {partition_device}")

            env = dict(os.environ)
            env["LVM_SUPPRESS_FD_WARNINGS"] = "1"
            output = vmdb.runcmd_chroot(
                chroot,
                ["lvs", "-o", "vg_name,devices", "--reportformat", "json"],
                env=env,
            )
            # example "lvs" output:
            #  {
            #    "report": [
            #        {
            #            "lv": [
            #                {"vg_name":"lvm_volgroup0", "devices":"/dev/mapper/loop1p6(1024)"},
            #                {"vg_name":"lvm_volgroup0", "devices":"/dev/mapper/loop1p6(0)"}
            #            ]
            #        }
            #    ]
            #  }
            report = json.loads(output.strip())
            logging.debug(f"lvs report: {report}")

            for report_item in report["report"]:
                logging.debug(f"report_item: {report_item}")

                for lv in report_item.get("lv", []):
                    logging.debug(f"lv: {lv}")
                    if lv.get("vg_name") == vgname:
                        devices = lv.get("devices", [])
                        if not isinstance(devices, list):
                            devices = [devices]
                        for device in devices:
                            logging.debug(f"device: {device}")
                            m = re.match(
                                r"^/dev/mapper/(?P<loop>.*)p\d+\(\d+\)$", device
                            )
                            if m is not None:
                                loop = m.group("loop")
                                return "/dev/{}".format(loop)

                            # Sometimes lvs gives us /dev/dm-X instead of a
                            # loop device, in which case we need to find the
                            # corresponding loop device ourselves
                            loop = self.get_loop_device_from_dm(device, chroot)
                            if loop is not None:
                                return loop

        raise Exception(
            "Do not understand partition device name {}".format(partition_device)
        )

    @staticmethod
    def get_loop_device_from_dm(device, chroot):
        # Returns the corresponding /dev/loopX device for a /dev/dm-Y device
        m = re.match(r"^/dev/(?P<mapped>dm-\d+)\(\d+\)$", device)
        if m is not None:
            mapped = m.group("mapped")
            logging.debug(f"mapped: {mapped}")

            dmsetup_out = vmdb.runcmd_chroot(
                chroot, ["dmsetup", "ls", "-o", "blkdevname"]
            )
            for line in dmsetup_out.decode().splitlines(keepends=False):
                m = re.match(rf"^(?P<loop>loop\d+)p\d+\s+\({mapped}\)$", line)
                if m is not None:
                    loop = m.group("loop")
                    return "/dev/{}".format(loop)

    def bind_mount_many(self, chroot, paths, state):
        for path in paths:
            self.mount(chroot, path, path, state, mount_opts=["--bind"])

    def mount(self, chroot, path, mount_point, state, mount_opts=None):
        chroot_path = self.chroot_path(chroot, mount_point)
        if os.path.ismount(chroot_path):
            logging.debug("already mounted: %s", chroot_path)
        else:
            if not os.path.exists(chroot_path):
                os.makedirs(chroot_path)

            if mount_opts is None:
                mount_opts = []

            vmdb.runcmd(["mount"] + mount_opts + [path, chroot_path])
            state.grub_mounts.append(chroot_path)

    def chroot_path(self, chroot, path):
        return os.path.normpath(os.path.join(chroot, "." + path))

    def install_package(self, chroot, package):
        env = os.environ.copy()
        env["DEBIAN_FRONTEND"] = "noninteractive"
        vmdb.runcmd_chroot(chroot, ["apt-get", "update"], env=env)
        vmdb.runcmd_chroot(
            chroot, ["apt-get", "-y", "--no-show-progress", "install", package], env=env
        )

    def set_grub_cmdline_config(self, chroot, kernel_params):
        param_string = " ".join(kernel_params)
        self.set_grub_default(
            chroot, "GRUB_CMDLINE_LINUX_DEFAULT", '"' + param_string + '"'
        )

    def set_grub_default(self, chroot, param, value):
        filename = self.chroot_path(chroot, "/etc/default/grub")
        newdefault = param + "=" + str(value) + "\n"

        found_param = False
        newcontents = ""
        with open(filename, "r+") as f:
            for line in f:
                if line.startswith(param + "="):
                    newcontents += newdefault
                    found_param = True
                elif line.startswith("#" + param + "="):
                    newcontents += line
                    newcontents += newdefault
                    found_param = True
                else:
                    newcontents += line
            if found_param:
                f.seek(0)
                f.write(newcontents)
                f.truncate()
            else:
                f.write(newdefault)

    def add_grub_serial_console(self, chroot):
        self.set_grub_default(chroot, "GRUB_TERMINAL", "serial")
        self.set_grub_default(
            chroot,
            "GRUB_SERIAL_COMMAND",
            '"serial ' '--speed=115200 --unit=0 --word=8 --parity=no --stop=1"',
        )

    def add_grub_crypto_disk(self, chroot):
        self.set_grub_default(chroot, "GRUB_ENABLE_CRYPTODISK", "y")

    def set_grub_timeout(self, chroot, timeout):
        self.set_grub_default(chroot, "GRUB_TIMEOUT", timeout)



# Copyright 2018  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import vmdb


class VgcreatePlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(VgcreateStepRunner())


class VgcreateStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {"vgcreate": str, "physical": []}

    def run(self, values, settings, state):
        vgname = self.get_vg(values)
        physical = self.get_pv(values, state)

        for phys in physical:
            vmdb.runcmd(["pvcreate", "-ff", "--yes", phys])
        vmdb.runcmd(["vgcreate", vgname] + physical)

    def teardown(self, values, settings, state):
        vgname = self.get_vg(values)
        vmdb.runcmd(["vgchange", "-an", vgname])

    def get_vg(self, values):
        return values["vgcreate"]

    def get_pv(self, values, state):
        return [state.tags.get_dev(tag) for tag in values["physical"]]



# Copyright 2018  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import unittest


import vmdb


class UnmountTests(unittest.TestCase):
    def setUp(self):
        self.mounts = ProcMounts()

    def unmount(self, what):
        vmdb.unmount(what, mounts=str(self.mounts), real_unmount=self.mounts.unmount)

    def test_raises_error_if_not_mounted(self):
        with self.assertRaises(vmdb.NotMounted):
            self.unmount("/foo")

    def test_unmounts_mounted_dir(self):
        self.mounts.mount("/dev/foo", "/foo")
        self.unmount("/foo")
        self.assertFalse(self.mounts.is_mounted("/foo"))

    def test_unmounts_mounted_dir_with_submounts(self):
        self.mounts.mount("/dev/foo", "/foo")
        self.mounts.mount("/dev/bar", "/foo/bar")
        self.unmount("/foo")
        self.assertFalse(self.mounts.is_mounted("/foo"))
        self.assertFalse(self.mounts.is_mounted("/foo/bar"))


class ProcMounts:
    def __init__(self):
        self.mounts = []

    def is_mounted(self, what):
        return any(what in mount for mount in self.mounts)

    def mount(self, device, point):
        self.mounts.append((device, point))

    def unmount(self, what):
        self.mounts = [mount for mount in self.mounts if what not in mount]

    def __str__(self):
        return "".join("{}\n".format(self.mount_line(mount)) for mount in self.mounts)

    def mount_line(self, mount):
        return "{} {} fstype options 0 0".format(mount[0], mount[1])



# Copyright 2017  Lars Wirzenius
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# =*= License: GPL-3+ =*=


import os

import vmdb


class SetPartFlagPlugin(vmdb.Plugin):
    def enable(self):
        self.app.step_runners.add(SetPartFlagStepRunner())


class SetPartFlagStepRunner(vmdb.StepRunnerInterface):
    def get_key_spec(self):
        return {
            "set_part_flag": str,
            "tag": str,
            "flag": str,
            "state": "enabled"
        }

    def run(self, values, settings, state):
        device = values["set_part_flag"]
        tag = values["tag"]
        flag = values["flag"]
        flag_state = values["state"]

        if flag_state == "enabled":
            flag_state = "on"
        elif flag_state == "disabled":
            flag_state = "off"
        else:
            raise SetPartFlagError('state must be "enabled" or "disabled"')

        tags_list = state.tags.get_tags()
        try:
            partition_number = tags_list.index(tag) + 1
        except IndexError:
            raise SetPartFlagError(f"Didn't find tag {tag} in {tags_list}")

        device = os.path.realpath(device)
        vmdb.runcmd([
            "parted", "-s", device, "--", "set",
            str(partition_number), flag, flag_state
        ])


class SetPartFlagError(Exception):
    pass
